/* BMP16.C

   Copyright (c)2002 Andrew Durdin. Email: andy@durdin.net

   You have permission to use this code in your own projects, on the conditions
   that leave this notice intact, and that you give proper credit in your final
   program. It'd be nice if you let me know what you do with it, also.
*/

#include <stdio.h>
#include <stdlib.h>

#include "bmp16.h"

#define PACKED  __attribute__((packed))

#define BMP_SIG			0x4D42
#define BI_RGB          0

typedef PACKED struct  BITMAPFILEHEADER {
	unsigned short bfType PACKED;			/* Must be 0x4D42 = "BM" (spec) */
	unsigned long  bfSize PACKED;			/* Size in bytes of BMP file */
	unsigned short bfReserved1 PACKED;		/* Must be 0 (spec) */
	unsigned short bfReserved2 PACKED;     /* Must be 0 (spec) */
	unsigned long  bfOffBits PACKED;		/* Offset in bytes to data */
} BITMAPFILEHEADER;

typedef PACKED struct BITMAPINFOHEADER {
	unsigned long biSize PACKED;			/* Size of structure: 40 */
	long biWidth PACKED;					/* Width of bitmap in pixels */
	long biHeight PACKED;					/* Negative heights not supported */
	unsigned short biPlanes PACKED;		/* Must be 1 (spec) */
	unsigned short biBitCount PACKED;		/* Only 4 supported */
	unsigned long biCompression PACKED;	/* Only BI_RGB supported */
	unsigned long biSizeImage PACKED;		/* Size in bytes of image */
	long biXPelsPerMeter PACKED;			/* Ignored - 0 */
	long biYPelsPerMeter PACKED;			/* Ignored - 0 */
	unsigned long biClrUsed PACKED;		/* Ignored - 0*/
	unsigned long biClrImportant PACKED;	/* Ignored - 0 */
} BITMAPINFOHEADER;

typedef PACKED struct RGBQUAD {
	unsigned char rgbBlue PACKED;
    unsigned char rgbGreen PACKED;
    unsigned char rgbRed PACKED;
    unsigned char rgbReserved PACKED;
} RGBQUAD;

void bmp_free( BITMAP16 *bmp ) {
	if( bmp ) {
    	if( bmp->lines )
        	free( bmp->lines );
        if( bmp->bits )
        	free( bmp->bits );
        free( bmp );
    }
}

BITMAP16 *bmp_load( char *fname ) {
	BITMAPFILEHEADER bfh;
    BITMAPINFOHEADER bih;
	BITMAP16 *bmp;
    FILE *fin;
    int y;

    /* Open the input picture */
    fin = fopen( fname, "rb" );
    if( !fin ) {
		return NULL;
    }

	/* Read the header and make sure it's a real BMP */
    fread( &bfh, sizeof( BITMAPFILEHEADER ), 1, fin );
    if( bfh.bfType != BMP_SIG ) {
    	fclose( fin );
        return NULL;
    }

    /* Make sure it's a format we can handle */
    fread( &bih, sizeof( BITMAPINFOHEADER ), 1, fin );
    if( bih.biPlanes != 1 || bih.biBitCount != 4 ||
			bih.biCompression != BI_RGB || bih.biHeight < 0 ) {
    	fclose( fin );
        return NULL;
    }

    /* Create a memory bitmap */
    bmp = bmp_create( bih.biWidth, bih.biHeight );
    if( !bmp ) {
       	fclose( fin );
        return NULL;
    }

	/* Now load the data into the buffer, one line at a time */
    fseek( fin, bfh.bfOffBits, SEEK_SET );
    for( y = bmp->h - 1; y >= 0; y-- )
        fread( bmp->lines[y], bmp->linew, 1, fin );

    /* Close the input file and return the bitmap pointer */
    fclose( fin );
	return bmp;
}

int bmp_save( BITMAP16 *bmp, char *fname ) {
	BITMAPFILEHEADER bfh;
    BITMAPINFOHEADER bih;
    RGBQUAD palette[16];
    FILE *fout;
	int i, y;

    if( !bmp )
    	return 0;

	if( bmp->planar ) {
		/**************************************************************/
		return 0;
	}


    /* Open the output picture */
    fout = fopen( fname, "wb" );
    if( !fout ) {
    	return 0;
    }

    /* Initialise the BITMAPFILEHEADER */
	bfh.bfType = BMP_SIG;
	bfh.bfSize = sizeof( BITMAPFILEHEADER ) + sizeof( BITMAPINFOHEADER ) +
		16 * sizeof( RGBQUAD ) + bmp->linew * bmp->h;
	bfh.bfReserved1 = 0;
	bfh.bfReserved2 = 0;
	bfh.bfOffBits = sizeof( BITMAPFILEHEADER ) + sizeof( BITMAPINFOHEADER ) +
		16 * sizeof( RGBQUAD );

    /* Initialise the BITMAPINFOHEADER */
	bih.biSize = sizeof( BITMAPINFOHEADER );
	bih.biWidth = bmp->w;
	bih.biHeight = bmp->h;
	bih.biPlanes = 1;
	bih.biBitCount = 4;
	bih.biCompression = BI_RGB;
	bih.biSizeImage = bmp->linew * bmp->h;
	bih.biXPelsPerMeter = 0;
	bih.biYPelsPerMeter = 0;
	bih.biClrUsed = 0;
	bih.biClrImportant = 0;

    /* Initialise the palette */
	for( i = 0; i < 16; i++ ) {
    	palette[i].rgbRed = ((i & 4) != 0) * 168 + ((i & 8) != 0) * 84;
    	palette[i].rgbGreen = ((i & 2) != 0) * 168 + ((i & 8) != 0) * 84;
    	palette[i].rgbBlue = ((i & 1) != 0) * 168 + ((i & 8) != 0) * 84;
    	palette[i].rgbReserved = 0;
    }
    /* Make palette entry 6 brown, not spew */
    palette[6].rgbGreen = 84;

	/* Write the headers and palette */
    fwrite( &bfh, sizeof( BITMAPFILEHEADER ), 1, fout );
    fwrite( &bih, sizeof( BITMAPINFOHEADER ), 1, fout );
    fwrite( palette, sizeof( RGBQUAD ), 16, fout );

	/* Now save the data from the buffer, one line at a time */
    for( y = bmp->h - 1; y >= 0; y-- )
        fwrite( bmp->lines[y], bmp->linew, 1, fout );

    /* Close the output file and return success */
    fclose( fout );
	return 1;
}

BITMAP16 *bmp_create( int w, int h ) {
	BITMAP16 *bmp;
    int y;

    /* Allocate the BITMAP structure */
    bmp = (BITMAP16 *)malloc( sizeof( BITMAP16 ) );
    if( !bmp ) {
        return NULL;
    }
    bmp->w = w;
    bmp->h = h;
    bmp->planar = 0;
    bmp->linew = ((bmp->w << 1) + 3) >> 2;	/* Allow for dword alignment */
    bmp->lines = NULL;
    bmp->bits = NULL;

    /* Allocate the pixel buffer */
    bmp->bits = (unsigned char *)malloc( bmp->h * bmp->linew *
    	sizeof( unsigned char ) );
    if( !bmp->bits ) {
    	bmp_free( bmp );
    	return NULL;
    }

    /* Allocate the lines array */
	bmp->lines = (unsigned char **)malloc( bmp->h *
		sizeof( unsigned char * ) );
    if( !bmp->lines ) {
    	bmp_free( bmp );
        return NULL;
    }
    for( y = 0; y < bmp->h; y++ )
    	bmp->lines[y] = bmp->bits + bmp->linew * y;

    /* Clear the bitmap */
    memset( bmp->bits, 0, bmp->linew * bmp->h );

    return bmp;
}

int bmp_getpixel( BITMAP16 *bmp, int x, int y ) {
	register int shift;
    if( bmp->planar || x < 0 || x >= bmp->w || y < 0 || y >= bmp->h )
    	return -1;

    shift = (x & 1) ? 0 : 4;
    return (bmp->lines[y][x / 2] >> shift) & 0x0F;
}

int bmp_putpixel( BITMAP16 *bmp, int x, int y, int c ) {
	register int mask, shift;
    if( bmp->planar || x < 0 || x >= bmp->w || y < 0 || y >= bmp->h )
    	return -1;

	shift = (x & 1) ? 0 : 4;
    mask = 0x0F << shift;
    bmp->lines[y][x / 2] &= ~mask;
    bmp->lines[y][x / 2] |= (c << shift) & mask;

    return c;
}

void bmp_rect( BITMAP16 *bmp, int x1, int y1, int x2, int y2, int c ) {
	int x, y;

//  	for( x = x1; x <= x2; x++ ) {
//      bmp_putpixel( bmp, x, y1, c );
//      bmp_putpixel( bmp, x, y2, c );
//   }
//  	for( y = y1; y <= y2; y++ ) {
//      bmp_putpixel( bmp, x1, y, c );
//      bmp_putpixel( bmp, x2, y, c );
//   }
	for( y = y1; y <= y2; y++ )
	   	for( x = x1; x <= x2; x++ )
	      	bmp_putpixel( bmp, x, y, c );
}

void bmp_unpack( BITMAP16 *bmp ) {
	/* Unpack a 4bpp bitmap into planes b, g, r, h */
	int unpackw;
	unsigned char **uplines;
	unsigned char *upbits;
	unsigned char *pix;
	int p, x, y, count;
	unsigned char byte;

	unpackw = (bmp->w + 7) / 8;

	upbits = malloc( unpackw * bmp->h * 4 * sizeof( unsigned char ) );
	uplines = malloc( bmp->h * 4 * sizeof( unsigned char * ) );

	for( p = 0; p < 4; p++ ) {
		for( y = 0; y < bmp->h; y++ ) {
			uplines[p * bmp->h + y] = upbits + (p * unpackw * bmp->h) +
				(y * unpackw);

			pix = uplines[p * bmp->h + y];
			count =	0;
			byte = 0;
			for( x = 0; x < bmp->w; x++ ) {
				byte <<= 1;
				byte |= ((bmp_getpixel( bmp, x, y ) & (1 << p)) ? 1 : 0);
				count++;
				if( count == 8 ) {
					/* output the byte */
					*pix++ = byte;
					count = 0;
					byte = 0;
				}
			}
			if( count != 0 ) {
				/* output the byte */
				*pix++ = byte;
			}
		}
	}

	free( bmp->lines );
	free( bmp->bits );

	bmp->bits = upbits;
	bmp->lines = uplines;

	bmp->linew = unpackw;
	bmp->planar = 1;
}

